use based::get_pg;
use cmd_lib::{FunResult, run_fun};
use owl::{prelude::*, query};
use serde::{Deserialize, Serialize};
use sqlx::FromRow;
use std::fs::read_dir;
use std::io::Read;
use std::ops::Deref;
use std::path::Path;
use std::process::Command;
use std::{fs, path::PathBuf};

use crate::PackageIndex;
use crate::builder::{BuildEnv, build};
use crate::git::{is_git, pull};

#[model]
pub struct IndexEntry {
    pub id: Id,
    pub repo: String,
    pub pkg: String,
    pub last_commit: Option<String>,
    pub last_pkgver: Option<String>,
}

impl IndexEntry {
    pub async fn all() -> Vec<Model<Self>> {
        query!(|_| true)
    }

    /// The path to the indexed pkgs build environment
    pub fn path(&self) -> PathBuf {
        std::path::Path::new("./repositories")
            .join(&self.repo)
            .join(&self.pkg)
    }

    pub fn build_dir(&self) -> String {
        format!("./repositories/{}/{}", self.repo, self.pkg)
    }
}

pub async fn rebuild_pkgs() {
    log::info!("Start checking for package rebuilds");

    for pkg in &IndexEntry::all().await {
        let mut should_rebuild = false;

        if is_git(pkg.read().path().to_str().unwrap()) {
            let git_commit = pull(pkg.read().path().to_str().unwrap()).unwrap_or_default();
            if let Some(old_commit) = &pkg.read().last_commit {
                if *old_commit != git_commit {
                    log::info!(
                        "Package {} / {} should rebuild: Commit {} -> {}",
                        pkg.read().repo,
                        pkg.read().pkg,
                        old_commit,
                        git_commit
                    );
                    should_rebuild = true;
                }
            } else {
                log::info!(
                    "Package {} / {} should rebuild: No last commit",
                    pkg.read().repo,
                    pkg.read().pkg
                );
                should_rebuild = true;
            }
        }

        let pkgver = get_pkgver(pkg.read().path().to_str().unwrap()).unwrap_or_default();

        if let Some(old_pkgver) = &pkg.read().last_pkgver {
            if *old_pkgver != pkgver {
                log::info!(
                    "Package {} / {} should rebuild: pkgver {} -> {}",
                    pkg.read().repo,
                    pkg.read().pkg,
                    old_pkgver,
                    pkgver
                );
                should_rebuild = true;
            }
        } else {
            log::info!(
                "Package {} / {} should rebuild: No last pkgver",
                pkg.read().repo,
                pkg.read().pkg
            );
            should_rebuild = true;
        }

        if should_rebuild {
            build(pkg).await;
        }
    }
    log::info!("Done checking for package rebuilds");
}

pub fn get_pkgver(path: &str) -> Option<String> {
    let out: FunResult = run_fun!(cd $path;pacco pkg info);

    if let Ok(out) = out {
        for line in out.lines() {
            let trimmed = line.trim();
            if let Some(pkgver) = trimmed.strip_prefix("pkgver = ") {
                return Some(pkgver.to_string());
            }
        }
    }

    None
}

pub fn find_package_files(dir: &str) -> Vec<String> {
    let output = Command::new("sh")
        .arg("-c")
        .arg(&format!("cd {dir};ls -1 *.pkg.tar.*"))
        .output()
        .expect("failed to execute process");

    let res = String::from_utf8_lossy(&output.stdout).to_string();
    res.split("\n")
        .filter_map(|x| {
            let x = x.trim();
            if x.is_empty() {
                None
            } else {
                if x.ends_with(".sig") {
                    return None;
                }
                Some(x.to_string())
            }
        })
        .collect()
}

pub fn find_pkgs(dir: &str) -> Vec<(String, Vec<u8>)> {
    let mut pkgs = Vec::new();

    let path = Path::new(dir);

    if let Ok(entries) = fs::read_dir(path) {
        for entry in entries.flatten() {
            let file_path = entry.path();
            if let Some(ext) = file_path.extension() {
                if ext == "zst" {
                    if let Some(file_name) = file_path.file_name().and_then(|n| n.to_str()) {
                        if file_name.ends_with(".tar.zst") {
                            match fs::File::open(&file_path) {
                                Ok(mut file) => {
                                    let mut content = Vec::new();
                                    if file.read_to_end(&mut content).is_ok() {
                                        pkgs.push((file_name.to_string(), content));
                                    }
                                }
                                Err(_) => continue,
                            }
                        }
                    }
                }
            }
        }
    }

    pkgs
}

pub async fn reindex_pkg(repo_dir: &Path) {
    log::info!("Indexing package builds");
    let mut dir_entries = read_dir(repo_dir).unwrap();
    let mut index = PackageIndex;

    while let Some(Ok(repo_entry)) = dir_entries.next() {
        if repo_entry.file_type().unwrap().is_dir() {
            let repo_name = repo_entry.file_name().into_string().unwrap();
            let mut pkg_entries = read_dir(repo_entry.path()).unwrap();

            while let Some(Ok(pkg_entry)) = pkg_entries.next() {
                if pkg_entry.file_type().unwrap().is_dir() {
                    let pkg_name = pkg_entry.file_name().into_string().unwrap();
                    log::info!("Found package {repo_name} / {pkg_name}");

                    index.insert_if_not_present(&repo_name, &pkg_name);
                }
            }
        }
    }

    log::info!("Done indexing package builds");
}
